/**********************************************************************
Copyright (c) 2011 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
   ...
**********************************************************************/
package org.datanucleus.store.mongodb.query;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.datanucleus.FetchPlan;
import org.datanucleus.metadata.AbstractClassMetaData;
import org.datanucleus.metadata.IdentityType;
import org.datanucleus.store.ExecutionContext;
import org.datanucleus.store.mongodb.MongoDBUtils;
import org.datanucleus.store.query.AbstractCandidateLazyLoadList;

import com.mongodb.DBCursor;
import com.mongodb.DBObject;

/**
 * Wrapper for lazy-loaded list of results from a MongoDB query.
 * Usage is as follows :-
 * <pre>
 * MongoDBCandidateList results = new MongoDBCandidateList(myCls, subclasses, ec, cacheType, fp, ignoreCache);
 * results.addCandidateCursor(cmd1, curs1);
 * results.addCandidateCursor(cmd2, curs2);
 * results.addCandidateCursor(cmd3, curs3);
 * ... use as a normal list now
 * results.close();
 * </pre>
 * Access to the list (and any iterator for it) will lazily load the candidate instances from the provided DBCursors.
 */
public class MongoDBCandidateList extends AbstractCandidateLazyLoadList
{
    List<DBCursor> cursors = new ArrayList<DBCursor>();

    int size = -1;

    FetchPlan fp;

    boolean ignoreCache;

    /**
     * @param cls
     * @param subclasses
     * @param ec
     * @param cacheType
     */
    public MongoDBCandidateList(Class cls, boolean subclasses, ExecutionContext ec, String cacheType,
            FetchPlan fp, boolean ignoreCache)
    {
        super(cls, subclasses, ec, cacheType);
        this.ignoreCache = ignoreCache;
        this.fp = fp;

        // Add placeholders for the cursors (populated by addCandidateCursor)
        for (int i=0;i<cmds.size();i++)
        {
            cursors.add(null);
        }
    }

    public void addCandidateCursor(AbstractClassMetaData cmd, DBCursor cursor)
    {
        int position = cmds.indexOf(cmd);
        cursors.set(position, cursor);
    }

    public void close()
    {
        Iterator<DBCursor> cursorIter = cursors.iterator();
        while (cursorIter.hasNext())
        {
            DBCursor cursor = cursorIter.next();
            cursor.close();
        }
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.query.AbstractLazyLoadList#retrieveObjectForIndex(int)
     */
    @Override
    protected Object retrieveObjectForIndex(int index)
    {
        Iterator<DBCursor> cursorIter = cursors.iterator();
        Iterator<AbstractClassMetaData> cmdIter = cmds.iterator();
        int first = 0;
        int last = -1;
        while (cursorIter.hasNext())
        {
            DBCursor cursor = cursorIter.next();
            AbstractClassMetaData cmd = cmdIter.next();
            last = first + cursor.size();
            if (index >= first && index < last)
            {
                // Found in this cursor
                // TODO This assumes we are always going forward continually. Needs to cater for jumping back
                int i= first + cursor.numSeen();
                while (cursor.hasNext())
                {
                    DBObject dbObject = cursor.next();
                    if (i == index)
                    {
                        int[] fpMembers = fp.getFetchPlanForClass(cmd).getMemberNumbers();
                        if (cmd.getIdentityType() == IdentityType.APPLICATION)
                        {
                            return MongoDBUtils.getObjectUsingApplicationIdForDBObject(dbObject, cmd, ec, ignoreCache,
                                fpMembers);
                        }
                        else if (cmd.getIdentityType() == IdentityType.DATASTORE)
                        {
                            return MongoDBUtils.getObjectUsingDatastoreIdForDBObject(dbObject, cmd, ec, ignoreCache,
                                fpMembers);
                        }
                    }
                    i++;
                }
            }
            else
            {
                first = last;
            }
        }

        return null;
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.query.AbstractLazyLoadList#getSize()
     */
    @Override
    protected int getSize()
    {
        if (size >= 0)
        {
            return size;
        }

        size = 0;
        Iterator<DBCursor> cursorIter = cursors.iterator();
        while (cursorIter.hasNext())
        {
            DBCursor cursor = cursorIter.next();
            size += cursor.size();
        }

        return size;
    }
}